import javafx.animation.Interpolator;
import javafx.animation.KeyFrame;
import javafx.animation.KeyValue;
import javafx.animation.Timeline;
import javafx.application.Application;
import javafx.scene.Group;
import javafx.scene.Scene;
import javafx.scene.input.KeyCode;
import javafx.scene.paint.Color;
import javafx.scene.shape.Circle;
import javafx.scene.shape.Line;
import javafx.scene.transform.Rotate;
import javafx.stage.Stage;
import javafx.util.Duration;

import java.util.ArrayList;
import java.util.List;

import static java.lang.Math.*;
import static java.util.stream.IntStream.range;

public class RollingCircles extends Application {

    private static final double ballRadius = 10.0;
    private static final int numberOfBalls = 8;
    private static final double RADIUS = 250.0;
    private static final double PERIOD = 2500.0; // rotation (rolling) period in millisec

    // mnemonics for ease of use and shortness of statements
    private final double R = RADIUS;
    private final double r = ballRadius;
    private final int N = numberOfBalls;
    private final double T = PERIOD;

    // the interpolator for small ball timeline to make it move
    // along its "spindle" in an oscillatory fashion (ie, accelerating
    // towards the centre, slowing down around turning points (at the edge)
    // -- like an oscillator
    private final Interpolator ip = new Interpolator() {
        @Override
        protected double curve(double t) {
            return interpolator(t);
        }
    };

    List<Timeline> timelines = new ArrayList<>();

    public static void main(String[] args) {
        launch(args);
    }

    private static double interpolator(double x) {
        return 0.5 * (1 - cos(PI * x));
    }

    @Override
    public void start(Stage primaryStage) throws Exception {
        Group root = new Group();
        Scene scene = new Scene(root, 2 * RADIUS, 2 * RADIUS);
        primaryStage.setTitle("Rolling or not?");
        primaryStage.setScene(scene);
        primaryStage.show();

        // big circle inside of which the "rolling" takes place
        Circle mainCircle = new Circle(R, R, R, Color.BLACK);
        root.getChildren().add(mainCircle);

        // drawing spindles
        range(0, numberOfBalls)
                .forEach(i -> {
                    Line line = new Line(
                            R * (1 + cos(PI * i / N)),
                            R * (1 + sin(PI * i / N)),
                            R * (1 - cos(PI * i / N)),
                            R * (1 - sin(PI * i / N)));

                    line.setStroke(Color.WHITE);
                    line.setStrokeWidth(1.0);
                    root.getChildren().add(line);
                });

        // callbacks to pause and resume rolling
        scene.onKeyPressedProperty().set(e -> {
            if (e.getCode() == KeyCode.P) {
                timelines.forEach(tl -> tl.pause());
            } else if (e.getCode() == KeyCode.R) {
                timelines.forEach(tl -> tl.play());
            }
        });

        // creating small balls and setting their timelines
        range(0, N).forEach(i -> {
            Timeline tl = makeBallWithTimeline(root, i);
            timelines.add(tl);
            tl.playFrom(Duration.millis((T / 2) * (1.0 * i / N)));
        });

    }

    private Timeline makeBallWithTimeline(Group group, int i) {
        Circle ball = new Circle(2 * R - r, R, r, Color.LEMONCHIFFON);
        group.getChildren().add(ball);
        Timeline tl = new Timeline();
        tl.setCycleCount(Timeline.INDEFINITE);
        tl.setAutoReverse(true);
        KeyValue kv = new KeyValue(ball.centerXProperty(), r, ip);
        KeyFrame kf = new KeyFrame(Duration.millis(T / 2), "half", kv);
        tl.getKeyFrames().add(kf);
        ball.getTransforms().add(new Rotate(180.0 * i / N, R, R));
        return tl;
    }
}

